from attr import attrs, attrib, validators, asdict


# IMPORTANT: Be careful when subclassing! Setting eq=False on a class whose base class has a non-default __hash__ method will not make attrs remove that __hash__ for you. The easiest way to reset __hash__ on a class is adding __hash__ = object.__hash__ in the class body.  # noqa
@attrs(eq=False, order=False, kw_only=True)
class BaseModel(object):
    def __eq__(self, target):
        if not isinstance(target, type(self)):
            return False
        hashed_self = self.__hash__()
        hashed_target = target.__hash__()
        if hashed_self is not None and hashed_target is not None:
            return hashed_self == hashed_target
        else:
            return False

    def __ne__(self, target):
        return not self.__eq__(target)

    def __lt__(self, target):
        return self.__hash__() < target.__hash__()

    def __le__(self, target):
        return self.__hash__() <= target.__hash__()

    def __gt__(self, target):
        return self.__hash__() > target.__hash__()

    def __ge__(self, target):
        return self.__hash__() >= target.__hash__()

    def __hash__(self):
        hashed_value = hash(self.id)
        return hashed_value

    def _ignore_base_attr(self, _attr, _val):
        return (
            False
            if _attr.name == "raw_data" or (_attr.name == "id" and not _val)
            else True
        )

    def to_dict(self):
        return asdict(self, filter=self._ignore_base_attr)

    id = attrib(factory=str)
    raw_data = attrib(factory=dict)


@attrs(eq=False)
class ErrorModel(BaseModel):
    __hash__ = BaseModel.__hash__

    code = attrib(default="0")
    message = attrib(default="UNEXPECTED ERROR.")


@attrs(eq=False)
class Credential(BaseModel):
    """
    Instantiate Credential model with dynamic attributes
    depending on various keys of the credential object.
    Args:
        credential (dict): The credential object.
    Example:
        cred = Credential(
            credential={"access_key_id": "abc", "secret_access_key": "*****"}
        )
        cred.access_key_id
        cred.secret_access_key
    """

    __hash__ = BaseModel.__hash__

    credential = attrib()

    @credential.validator
    def is_valid_credential(self, attribute, value):
        # Set credential key, value as model attribute dynamically.
        if isinstance(value, dict):
            for k, v in value.items():
                setattr(self, k, v)
        else:
            raise TypeError(f"The credential must be a dictionary, not '{type(value)}'")


@attrs(eq=False)
class Location(BaseModel):
    def __hash__(self):
        return hash(f"{self.lat},{self.lon}")

    lat = attrib(converter=float)
    lon = attrib(converter=float)


@attrs(eq=False)
class Temperature(BaseModel):
    __hash__ = BaseModel.__hash__

    temp_avg = attrib(default=-999, converter=float)
    temp_min = attrib(default=-999, converter=float)
    temp_max = attrib(default=-999, converter=float)


@attrs(eq=False)
class Wind(BaseModel):
    __hash__ = BaseModel.__hash__

    speed = attrib(converter=float)
    degree = attrib(converter=int)


@attrs(eq=False)
class Station(Location):
    def __hash__(self):
        hashed_value = hash(self.name)
        return hashed_value

    name = attrib(converter=str)


@attrs(eq=False)
class WeatherPoint(BaseModel):
    def __hash__(self):
        hashed_value = hash(self.station)
        return hashed_value

    station = attrib(validator=[validators.instance_of(Station)])
    status = attrib(converter=str)
    description = attrib(converter=str)
    icon = attrib(converter=str)
    time = attrib(converter=int)
    temp = attrib(validator=[validators.instance_of(Temperature)])
    pressure = attrib(converter=float)
    humidity = attrib(converter=float)
    rain = attrib(converter=float)
    wind = attrib(validator=[validators.instance_of(Wind)])
    clouds = attrib(converter=int)
    snow = attrib(converter=float)



import functools
import os
import re
import requests
import time
from random import randint
from requests import Session
from requests.exceptions import HTTPError
from requests.packages import urllib3

from .logger import get_logger

logger = get_logger()

RETRY_TIMES = 3
BACKOFF_FACTOR = 1
STATUS_TO_RETRY = (429, 500, 502, 503, 504)
DEFAULT_TIMEOUT = (30, 120)
URL_PATH_PATTERN = re.compile("([^/]).*")
DOMAIN_PATTERN = re.compile(r"^((?!-)[A-Za-z0-9-]{1,63}(?<!-)\.)+[A-Za-z]{2,63}$")


class HTTP_METHOD:
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


class RetryException(Exception):
    """Specified exception to retry request."""

    def __init__(self, message):
        super().__init__(message)


def is_ssl_verification():
    return False if os.environ.get("CURL_CA_BUNDLE") == "" else True


def retry(*exceptions):
    """Decorate an async function to execute it a few times before giving up.
    Hopes that problem is resolved by another side shortly.
    Args:
        exceptions (tuple): The exceptions expected during function execution
    """

    def accumulate_cooldown(retry_times, backoff_factor=None):
        backoff_factor = backoff_factor or BACKOFF_FACTOR
        return backoff_factor * (2 ** retry_times) + (randint(1, 999)) * 0.001

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _count = 0
            _errors = None
            while _count < RETRY_TIMES:
                try:
                    return func(*args, **kwargs)
                except exceptions as err:
                    _errors = err
                    if getattr(_errors, "response", None) is not None:
                        if err.response.status_code not in STATUS_TO_RETRY:
                            break

                    _count += 1
                    try:
                        retry_after = int(err.response.headers.get("Retry-After", 0))
                    except Exception:
                        retry_after = accumulate_cooldown(
                            _count, kwargs.get("backoff_factor")
                        )

                    logger.debug(f"Retry attempt {_count} after {retry_after} seconds")
                    time.sleep(retry_after)

            if _errors:
                raise _errors

        return wrapper

    return decorator


@retry(RetryException, HTTPError)
def retry_request(
    method,
    url,
    headers=None,
    files=None,
    data=None,
    json=None,
    params=None,
    auth=None,
    cookies=None,
    hooks=None,
    timeout=DEFAULT_TIMEOUT,
    verify=None,
    session=None,
    proxies=None,
    jsonify=True,
    object_hook=None,
    **kwargs,
):
    """
    Args:
        method: HTTP request method.
        url: HTTP request url.
        session: The session is used to send HTTP request.
        proxies: The proxy configs are applied to session.
        timeout: The specified timeout for HTTP response.
        jsonify: By default it is True, return response content as dictionary.
            If False, return HTTP response object.
        object_hook: The hook is used for json.loads if jsonify is True.
        error_handler: The function is used to handle raised exception and accept
            the exception object(maybe None) as first args of function.
    """

    session = session or Session()
    if isinstance(proxies, dict):
        session.proxies.update(proxies)

    req = requests.Request(
        method=method,
        url=url,
        headers=headers,
        files=files,
        data=data,
        json=json,
        params=params,
        auth=auth,
        cookies=cookies,
        hooks=hooks,
    ).prepare()
    resp_data = None
    verify = verify if isinstance(verify, bool) else is_ssl_verification()
    if verify is False:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    _start_time = time.time()
    _target = f"{method.upper()} {url}"
    logger.debug(f"HTTP request start, {_target}")
    resp = session.send(req, timeout=timeout, verify=verify, **kwargs)
    resp.raise_for_status()
    _dt = time.time() - _start_time
    logger.debug(
        f"HTTP request finish in {_dt:.3f} seconds, {_target}"
    )
    resp_data = resp.json(object_hook=object_hook) if jsonify else resp
    return resp_data


def urljoin(*paths):
    """
    urljoin(*args): Concatenate any number of strings as url path and handle for the
        path which starts with "/".
    Example: urljoin("https://www.pentium.network/", "/users", "me/")
        -> "https://www.pentium.network/users/me"
    """

    _paths = [str(a) for a in paths]
    _paths = (
        _match.group() for _match in map(URL_PATH_PATTERN.search, _paths) if _match
    )
    url_path = os.path.join(*_paths)
    return url_path


def chunks(lst, n):
    """Generator yield n-sized chunks from a list."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]



def humidity_from_dewpoint(fa, fdp):
    """Simple approximation of relative humidity formular

    Args:
        fa (float): The temperature of air in fahrenheit
        fdp (float): The temperature of dew point in fahrenheit

    Returns:
        float: The relative humidity
    """

    relative_humidity = 100 - 5 * (5 / 9) * (fa - fdp)
    return relative_humidity


def kelvin_from_fahrenheit(f):
    return (f + 459.67) * (5 / 9)


def fahrenheit_from_kelvin(k):
    return k * (9 / 5) - 459.67


def celsius_from_fahrenheit(f):
    return (5 / 9) * (f - 32)


def fahrenheit_from_celsius(c):
    return c * (9 / 5) + 32


def mps_to_kph(v):
    return v / (1000 / 3600)


def kph_to_mps(v):
    return v * (1000 / 3600)



import statistics
from copy import deepcopy

from wxkit.services import WxService
from wxkit.models import ErrorModel, Wind, Temperature, Station, WeatherPoint
from wxkit.utils.httplib import retry_request, urljoin, HTTP_METHOD


def strip_data(obj):
    """
    For some data, eg. rain, snow and clouds, the data structure of forecast weather is
    different from current weather.
    In order unified the difference, strip the data without the redundant key.
    """

    keywords = ("1h", "3h")
    if isinstance(obj, dict):
        for kw in keywords:
            try:
                val = obj[kw]
                return val
            except KeyError:
                pass

    return obj


class OpenweatherWxService(WxService):
    _CREDENTIAL_KEY = "appid"
    ENDPOINT = "https://api.openweathermap.org/data/2.5"
    PATH_CURRENT_WEATHER = ("weather",)
    PATH_FORECAST_3_HOURLY = ("forecast",)
    PATH_FORECAST_HOURLY = ("forecast", "hourly")
    FORECAST_PERIOD = (1, 3, 24)

    def __init__(self, credential):
        super().__init__(credential)

    def auth_params(self, **params):
        p = dict(appid=self.credential.appid)
        p.update(params)
        return p

    def _modelize(self, data):
        coord = data.get("coord", {})
        name = data.get("name", "")
        station = Station(lat=coord.get("lat", 0), lon=coord.get("lon", 0), name=name)
        temperature = Temperature(
            temp_avg=data.get("main", {}).get("temp", 0),
            temp_min=data.get("main", {}).get("temp_min", 0),
            temp_max=data.get("main", {}).get("temp_max", 0),
        )
        wind = Wind(
            speed=data.get("wind", {}).get("speed", 0),
            degree=data.get("wind", {}).get("deg", 0),
        )
        try:
            wx = data["weather"][0]
            status = wx.get("main", "")
            description = wx.get("description", "")
            icon = wx.get("icon", "")
        except Exception:
            status = ""
            description = ""
            icon = ""

        pressure = data.get("main", {}).get("pressure", 0)
        humidity = data.get("main", {}).get("humidity", 0)
        rain = data.get("rain", 0)
        clouds = data.get("clouds", {}).get("all", 0)
        snow = data.get("snow", 0)
        time = data.get("dt", 0)
        weather_point = WeatherPoint(
            station=station,
            status=status,
            description=description,
            icon=icon,
            temp=temperature,
            pressure=pressure,
            humidity=humidity,
            rain=rain,
            wind=wind,
            clouds=clouds,
            snow=snow,
            time=time,
            raw_data=data,
        )
        return weather_point

    def get_current_weather_by_location(self, location):
        url = urljoin(self.ENDPOINT, *self.PATH_CURRENT_WEATHER)
        params = self.auth_params(lat=location.lat, lon=location.lon, units="metric")
        try:
            resp_data = retry_request(
                HTTP_METHOD.GET, url, params=params, object_hook=strip_data
            )
        except Exception as e:
            return self.handle_error(e)
        return self._modelize(resp_data)

    def get_forecast_weather_by_location(self, location, period=3):
        if period not in self.FORECAST_PERIOD:
            return self.handle_error(f"Missing weather forecast per {period} hours")

        period_alias = (
            self.PATH_FORECAST_HOURLY if period == 1 else self.PATH_FORECAST_3_HOURLY
        )
        url = urljoin(self.ENDPOINT, *period_alias)
        params = self.auth_params(lat=location.lat, lon=location.lon, units="metric")
        try:
            resp_data = retry_request(
                HTTP_METHOD.GET, url, params=params, object_hook=strip_data
            )
        except Exception as e:
            return self.handle_error(e)

        coord = resp_data.get("city", {}).get("coord", {})
        name = resp_data.get("city", {}).get("name", "")
        data_collection = []
        for data in resp_data.get("list"):
            data.update(coord=coord, name=name)
            data_collection.append(self._modelize(data))

        # NOTE: Resampling daily weather forecast from per 3 hours.
        if period == 24:
            data_collection = self.resample_daily_wxs(data_collection)

        return data_collection

    def resample_daily_wxs(self, wxs):
        ONE_DAY_HOUR = 24
        delta_hour = (wxs[1].time - wxs[0].time) / 3600
        samples = int(ONE_DAY_HOUR / delta_hour)
        r_wxs = []
        for i in range(0, len(wxs), samples):
            temp_avg = []
            temp_min = []
            temp_max = []
            for wx in wxs[slice(i, (i + 1) * samples)]:
                temp_avg.append(wx.temp.temp_avg)
                temp_min.append(wx.temp.temp_min)
                temp_max.append(wx.temp.temp_max)

            r_wx = deepcopy(wxs[i])
            r_wx.temp.temp_avg = round(statistics.mean(temp_avg), 2)
            r_wx.temp.temp_min = round(statistics.mean(temp_min), 2)
            r_wx.temp.temp_max = round(statistics.mean(temp_max), 2)
            r_wxs.append(r_wx)

        return r_wxs

    def handle_error(self, exception):
        error = ErrorModel()
        if getattr(exception, "response", None) is not None:
            try:
                # Get error context from HTTP error if response.
                response = exception.response.json()
                error.code = response["cod"]
                error.message = response["message"]
                error.raw_data = response
            except Exception:
                error.code = exception.response.status_code
                error.message = exception.response.text
        else:
            error.message = str(exception)

        return error
